// var debug = require('debug')('holodex:agent:model')
var createGraph = require('ngraph.graph')
var Model = require('base/model')
var relationships = require('relationship/store')
var roles = require('role/store')
var store = require('base/store')
var keys = require('lodash.keys')

module.exports = Model.extend({
  typeName: 'Agent',
  collectionName: 'agents',

  props: {
    name: 'string',
    image: 'url'
  },

  initialize: function (attrs, opts) {
    // debug('initialize', this)
    Model.prototype.initialize.apply(this, arguments)

    this._roles = roles
      .addDynamicView(this.getId() + '/roles')
      .applyFind({
        agent: this.getId()
      })

    this._contextRelationships = relationships
      .addDynamicView(this.getId() + '/contextRelationships')
      .applyFind({
        context: this.getId()
      })
  },

  derived: {
    roles: {
      cache: false,
      fn: function () {
        return this._roles.data()
          .map(function (role) {
            return store.findById(role['@id'])
          })
      }
    },

    contextRelationships: {
      cache: false,
      fn: function () {
        return this._contextRelationships.data()
          .map(function (relationship) {
            return store.findById(relationship['@id'])
          })
      }
    },

    relationships: {
      cache: false,
      fn: function () {
        // debug('roles', this.roles)
        return this.roles.map(function (role) {
          return role.relationship
        })
          .filter(function (i) { return i })
          .concat(
            this._contextRelationships.data()
              .map(function (relationship) {
                return store.findById(relationship['@id'])
              })
        )
      }
    },

    relationshipGraph: {
      cache: false,
      fn: function () {
        // debug('relationships', this.relationship)

        var graph = createGraph()
        graph.beginUpdate()

        this.relationships.forEach(function (relationship) {
          // debug('relationship', relationship)
          // add agents to graph (if not already)
          relationship.agents.forEach(function (agent) {
            // debug('agent', agent)
            if (agent && !graph.getNode(agent.getId())) { // ?
              graph.addNode(agent.getId(), agent)
            }
          })

          // add role as link between agents
          relationship.roles.forEach(function (sourceRole) {
            relationship.roles.forEach(function (targetRole) {
              if (!sourceRole || !targetRole || sourceRole === targetRole) {
                return
              }
              if (!sourceRole.agent || !targetRole.agent) { return }
              var data = { sourceRole: sourceRole }
              sourceRole.relationship.type.linkTypes.forEach(function (lt) {
                if (lt.source.getId() === sourceRole.type.getId()) {
                  data.linkType = lt
                }
              })

              graph.addLink(sourceRole.agent.getId(), targetRole.agent.getId(), data)
            })
          })
        })

        graph.endUpdate()
        // debug('relationshipGraph ->', graph)
        return graph
      }
    },

    relsByType: {
      cache: false,
      deps: ['relationships'],
      fn: function () {
        var relsByType = {}
        this.relationships.forEach(function (relationship) {
          var relTypeId = relationship.type.getId()
          if (relsByType[relTypeId] == null) {
            relsByType[relTypeId] = []
          }
          relsByType[relTypeId].push(relationship)
        })
        return relsByType
      }
    },

    rolesByType: {
      cache: false,
      deps: ['roles'],
      fn: function () {
        var rolesByType = {}
        this.roles.forEach(function (role) {
          var roleTypeId = role.type.getId()
          if (rolesByType[roleTypeId] == null) {
            rolesByType[roleTypeId] = []
          }
          rolesByType[roleTypeId].push(role)
        })
        return rolesByType
      }
    },

    linkTypesByDirection: {
      cache: false,
      deps: ['rolesByType'],
      fn: function () {
        var rolesByType = this.rolesByType
        var linkTypesByDirection = { asSource: [], asTarget: [] }
        keys(rolesByType).forEach(function (roleId) {
          var role = rolesByType[roleId][0]
          role.relationship.type.linkTypes.forEach(function (linkType) {
            if (linkType.source.getId() === role.type.getId()) {
              linkTypesByDirection.asSource.push(linkType)
            }
            if (linkType.target.getId() === role.type.getId()) {
              linkTypesByDirection.asTarget.push(linkType)
            }
          })
        })

        return linkTypesByDirection
      }
    },

    pairs: {
      cache: false,
      deps: ['relationships'],
      fn: function () {
        var agentId = this.getId()
        var pairs = []

        this.roles.forEach(function (role) {
          if (role.pair.agent && role.pair.agent.getId() !== agentId) {
            pairs.push(role.pair)
          }
        })

        return pairs
      }
    },

    pairRoleTypes: {
      cache: false,
      deps: ['relationships'],
      fn: function () {
        var agentId = this.getId()
        var types = {}

        this.relationships.forEach(function (relationship) {
          relationship.roles.forEach(function (role) {
            if (
              role.agent &&
              role.agent.getId() !== agentId &&
              !types[role.type.getId()]
            ) {
              types[role.type.getId()] = role.type
            }
          })
        })

        return types
      }
    }
  }
})
